from django.test import TestCase, override_settings
from django.urls import reverse

from iapauth.middleware import StubAuthenticator


class MiddlewareTest(TestCase):
    def test_with_no_header(self):
        """we haven't set a JWT header, so middleware shouldn't be involved"""
        r = self.client.get(reverse("testview"))
        self.assertEqual(r.status_code, 200)
        self.assertTrue("current user: Anonymous" in str(r.content))
        self.assertTrue("is_authenticated: False" in str(r.content))

    @override_settings(JWT_AUDIENCE="/projects/foo")
    def test_invalid_jwt(self):
        headers = {"HTTP_X_GOOG_IAP_JWT_ASSERTION": "totally not a legit JWT"}
        r = self.client.get(reverse("testview"), **headers)
        # invalid JWT on a non-protected view is fine
        self.assertEqual(r.status_code, 200)
        # but we should definitely be logged out
        self.assertTrue("current user: Anonymous" in str(r.content))
        self.assertTrue("is_authenticated: False" in str(r.content))

    @override_settings(JWT_AUDIENCE="/projects/foo")
    def test_invalid_jwt_protected_view(self):
        headers = {"HTTP_X_GOOG_IAP_JWT_ASSERTION": "totally not a legit JWT"}
        r = self.client.get(reverse("protected"), **headers)
        # shouldn't allow us in there
        self.assertEqual(r.status_code, 302)

    @override_settings(
        JWT_AUDIENCE="/projects/foo",
        IAPAUTH_JWT_AUTHENTICATOR=StubAuthenticator(
            True, "testuser@example.com", "example.com"
        ),
    )
    def test_allow_regular_user(self):
        # we have a stubbed out authenticator that will pass
        headers = {"HTTP_X_GOOG_IAP_JWT_ASSERTION": "totally not a legit JWT"}
        r = self.client.get(reverse("testview"), **headers)
        self.assertEqual(r.status_code, 200)
        self.assertTrue("current user: testuser" in str(r.content))
        self.assertTrue("is_authenticated: True" in str(r.content))
        # should not be staff or superuser
        self.assertTrue("is_staff: False" in str(r.content))
        self.assertTrue("is_superuser: False" in str(r.content))

    @override_settings(
        JWT_AUDIENCE="/projects/foo",
        IAPAUTH_JWT_AUTHENTICATOR=StubAuthenticator(
            True, "testuser@example.com", "example.com"
        ),
    )
    def test_allow_regular_user_protected_view(self):
        # should also be able to get in with @login_required view
        headers = {"HTTP_X_GOOG_IAP_JWT_ASSERTION": "totally not a legit JWT"}
        r = self.client.get(reverse("protected"), **headers)
        self.assertEqual(r.status_code, 200)
        self.assertTrue("current user: testuser" in str(r.content))
        self.assertTrue("is_authenticated: True" in str(r.content))

    @override_settings(
        JWT_AUDIENCE="/projects/foo",
        IAPAUTH_JWT_AUTHENTICATOR=StubAuthenticator(
            True, "testuser@example.com", "example.com"
        ),
        IAP_JWT_ADMIN_DOMAIN="example.com",
    )
    def test_superuser_mapping(self):
        # we have a stubbed out authenticator that will pass
        headers = {"HTTP_X_GOOG_IAP_JWT_ASSERTION": "totally not a legit JWT"}
        r = self.client.get(reverse("testview"), **headers)
        self.assertEqual(r.status_code, 200)
        self.assertTrue("current user: testuser" in str(r.content))
        self.assertTrue("is_authenticated: True" in str(r.content))
        # should be staff and superuser
        self.assertTrue("is_staff: True" in str(r.content))
        self.assertTrue("is_superuser: True" in str(r.content))

    @override_settings(
        JWT_AUDIENCE="/projects/foo",
        IAPAUTH_JWT_AUTHENTICATOR=StubAuthenticator(
            True, "testuser@example.com", "example.com"
        ),
    )
    def test_login_twice_valid(self):
        headers = {"HTTP_X_GOOG_IAP_JWT_ASSERTION": "totally not a legit JWT"}
        r = self.client.get(reverse("testview"), **headers)
        self.assertEqual(r.status_code, 200)
        self.assertTrue("current user: testuser" in str(r.content))
        self.assertTrue("is_authenticated: True" in str(r.content))
        # now we access it again and it should still work
        r = self.client.get(reverse("testview"), **headers)
        self.assertEqual(r.status_code, 200)
        self.assertTrue("current user: testuser" in str(r.content))
        self.assertTrue("is_authenticated: True" in str(r.content))

    @override_settings(
        JWT_AUDIENCE="/projects/foo",
        IAPAUTH_JWT_AUTHENTICATOR=StubAuthenticator(
            True, "testuser@example.com", "example.com"
        ),
    )
    def test_login_failure_logs_user_out(self):
        headers = {"HTTP_X_GOOG_IAP_JWT_ASSERTION": "totally not a legit JWT"}
        r = self.client.get(reverse("testview"), **headers)
        self.assertEqual(r.status_code, 200)
        self.assertTrue("current user: testuser" in str(r.content))
        self.assertTrue("is_authenticated: True" in str(r.content))
        # now we access it again, but this time the authentication fails
        with self.settings(
            IAPAUTH_JWT_AUTHENTICATOR=StubAuthenticator(False, None, None),
        ):
            r = self.client.get(reverse("testview"), **headers)
            self.assertEqual(r.status_code, 200)
            # should be logged out
            self.assertTrue("current user: Anonymous" in str(r.content))
            self.assertTrue("is_authenticated: False" in str(r.content))

    @override_settings(
        JWT_AUDIENCE="/projects/foo",
        IAPAUTH_JWT_AUTHENTICATOR=StubAuthenticator(
            True, "testuser@example.com", "example.com"
        ),
    )
    def test_user_mismatch(self):
        """test the case where they are already logged in as one user, then authenticate
        as a different user"""
        headers = {"HTTP_X_GOOG_IAP_JWT_ASSERTION": "totally not a legit JWT"}
        r = self.client.get(reverse("testview"), **headers)
        self.assertEqual(r.status_code, 200)
        self.assertTrue("current user: testuser" in str(r.content))
        self.assertTrue("is_authenticated: True" in str(r.content))
        # now we access it again, but this time the authentication fails
        with self.settings(
            IAPAUTH_JWT_AUTHENTICATOR=StubAuthenticator(
                True, "second@example.com", "example.com"
            )
        ):
            r = self.client.get(reverse("testview"), **headers)
            self.assertEqual(r.status_code, 200)
            # should be logged out of the first one and logged in as the second
            self.assertTrue("current user: second" in str(r.content))
            self.assertTrue("is_authenticated: True" in str(r.content))



import os

import jose
import requests
from django.conf import settings
from django.contrib import auth
from django.contrib.auth import load_backend
from django.contrib.auth.backends import RemoteUserBackend
from django.core.exceptions import ImproperlyConfigured
from django.utils.deprecation import MiddlewareMixin
from jose import jwt


KEYS = None  # Cached public keys for verification
AUDIENCE = None  # Cached value requiring information from metadata server


# Google publishes the public keys needed to verify a JWT. Save them in KEYS.
def keys():
    global KEYS

    if KEYS is None:
        resp = requests.get("https://www.gstatic.com/iap/verify/public_key")
        KEYS = resp.json()

    return KEYS


# Returns the JWT "audience" that should be in the assertion
def gcp_jwt_audience():
    global AUDIENCE

    if AUDIENCE is None:
        project_id = os.getenv("GOOGLE_CLOUD_PROJECT", None)

        endpoint = "http://metadata.google.internal"
        path = "/computeMetadata/v1/project/numeric-project-id"
        response = requests.get(
            "{}/{}".format(endpoint, path), headers={"Metadata-Flavor": "Google"}
        )
        project_number = response.json()

        AUDIENCE = "/projects/{}/apps/{}".format(project_number, project_id)

    return AUDIENCE


class JWTAuthenticator(object):
    def authenticate(self, jwt_token, audience):
        try:
            info = jwt.decode(
                jwt_token,
                keys(),
                algorithms=["ES256"],
                audience=audience,
            )
            return (True, info["email"], info["hd"])
        except jose.exceptions.JOSEError as e:
            # log it
            print("bad JWT: {}".format(e))
        return (False, None, None)


class StubAuthenticator(object):
    """ stub for testing """
    def __init__(self, status=True, email="foo@example.com", domain="example.com"):
        self.response = (status, email, domain)

    def authenticate(self, jwt_token, audience):
        return self.response


class IAPJWTAuthMiddleware(MiddlewareMixin):
    def process_request(self, request):
        # AuthenticationMiddleware is required so that request.user exists.
        if not hasattr(request, "user"):
            raise ImproperlyConfigured(
                "The Django remote user auth middleware requires the"
                " authentication middleware to be installed.  Edit your"
                " MIDDLEWARE setting to insert"
                " 'django.contrib.auth.middleware.AuthenticationMiddleware'"
                " before the RemoteUserMiddleware class."
            )
        request.jwt_user_email = None
        request.jwt_domain = None
        request.jwt_authenticated = False

        jwt_token = request.META.get("HTTP_X_GOOG_IAP_JWT_ASSERTION", None)

        if not jwt_token:
            # no token. We have no responsibility here
            return
        if hasattr(settings, "JWT_AUDIENCE"):
            audience = settings.JWT_AUDIENCE
        else:
            audience = gcp_jwt_audience()

        # we can override the authenticator for testing
        authenticator = JWTAuthenticator()
        if hasattr(settings, "IAPAUTH_JWT_AUTHENTICATOR"):
            authenticator = settings.IAPAUTH_JWT_AUTHENTICATOR

        authenticated, email, domain = authenticator.authenticate(jwt_token, audience)
        if authenticated:
            request.jwt_user_email = email
            request.jwt_domain = domain
            request.jwt_authenticated = True
        else:
            # they tried to use a bad JWT. Make sure they
            # are logged out.
            self._ensure_logged_out(request)
            return

        username = self.clean_username(request.jwt_user_email, request)
        # If the user is already authenticated and that user is the user we are
        # getting passed in the headers, then the correct user is already
        # persisted in the session and we don't need to continue.
        if request.user.is_authenticated:
            if request.user.get_username() == username:
                return
            else:
                # An authenticated user is associated with the request, but
                # it does not match the authorized user in the header.
                self._remove_invalid_user(request)

        # We are seeing this user for the first time in this session, attempt
        # to authenticate the user.
        user = auth.authenticate(request, remote_user=username)
        if user:
            # User is valid.  Set request.user and persist user in the session
            # by logging the user in.
            request.user = user
            auth.login(request, user)

    def _ensure_logged_out(self, request):
        request.jwt_authenticated = False
        request.jwt_domain = None
        request.jwt_user_email = None
        if request.user.is_authenticated:
            self._remove_invalid_user(request)

    def clean_username(self, username, request):
        return username.split("@")[0]

    def _remove_invalid_user(self, request):
        """
        Remove the current authenticated user in the request which is invalid
        but only if the user is authenticated via the RemoteUserBackend.
        """
        try:
            stored_backend = load_backend(
                request.session.get(auth.BACKEND_SESSION_KEY, "")
            )
        except ImportError:
            # backend failed to load
            auth.logout(request)
        else:
            if isinstance(stored_backend, RemoteUserBackend):
                auth.logout(request)



from django.conf import settings
from django.contrib.auth.backends import RemoteUserBackend


# this expects to be used with IAPJWTAuthMiddleware
class IAPJWTUserBackend(RemoteUserBackend):
    def configure_user(self, request, user):
        """
        Configure a user after creation and return the updated user.
        """
        user.email = request.jwt_user_email
        user.set_unusable_password()
        domain = request.jwt_domain
        if hasattr(settings, "IAP_JWT_ADMIN_DOMAIN"):
            if domain == settings.IAP_JWT_ADMIN_DOMAIN:
                user.is_superuser = True
                user.is_staff = True
        user.save()
        return user



from django.contrib.auth.decorators import login_required
from django.http import HttpResponse


def testview(request):
    output = """
    current user: {}
    is_authenticated: {}
    is_staff: {}
    is_superuser: {}
    """.format(
        request.user,
        request.user.is_authenticated,
        request.user.is_staff,
        request.user.is_superuser,
    )
    return HttpResponse(output)


@login_required
def protected(request):
    output = """
    current user: {}
    is_authenticated: {}
    is_staff: {}
    is_superuser: {}
    """.format(
        request.user,
        request.user.is_authenticated,
        request.user.is_staff,
        request.user.is_superuser,
    )
    return HttpResponse(output)
